//
// Copyright (c) 2016-2020 CNRS INRIA
//

#ifndef __pinocchio_python_parser_python_hpp__
#define __pinocchio_python_parser_python_hpp__

#include "pinocchio/macros.hpp"

// clang-format off
PINOCCHIO_PRAGMA_DEPRECATED_HEADER(pinocchio/bindings/python/parsers/python.hpp, pinocchio/parsers/python.hpp)
// clang-format on

#include "pinocchio/parsers/python.hpp"

#endif // ifndef __pinocchio_python_parser_python_hpp__
