//
// Copyright (c) 2024 INRIA
//
//
#ifndef __pinocchio_python_collision_hpp__
#define __pinocchio_python_collision_hpp__

namespace pinocchio
{
  namespace python
  {

    void exposeBroadphaseCallbacks();
    void exposeBroadphase();

  } // namespace python
} // namespace pinocchio

#endif // ifndef __pinocchio_python_collision_hpp__
