//
// Copyright (c) 2019 INRIA
//

#ifndef __pinocchio_math_casadi_hpp__
#define __pinocchio_math_casadi_hpp__

#include "pinocchio/macros.hpp"

// clang-format off
PINOCCHIO_PRAGMA_DEPRECATED_HEADER(pinocchio/math/casadi.hpp, pinocchio/autodiff/casadi.hpp)
// clang-format on

#include "pinocchio/autodiff/casadi.hpp"

#endif // #ifndef __pinocchio_math_casadi_hpp__
