//
// Copyright (c) 2021 INRIA
//

#ifndef __pinocchio_algorithm_contact_hpp__
#define __pinocchio_algorithm_contact_hpp__

#include "pinocchio/algorithm/contact-info.hpp"
#include "pinocchio/algorithm/contact-cholesky.hpp"
#include "pinocchio/algorithm/contact-dynamics.hpp"
#include "pinocchio/algorithm/contact-dynamics-derivatives.hpp"
#include "pinocchio/algorithm/contact-jacobian.hpp"

namespace pinocchio
{

} // namespace pinocchio

#include "pinocchio/algorithm/contact.hxx"

#endif // ifndef __pinocchio_algorithm_contact_hpp__
