//
// Copyright (c) 2024 INRIA
//

#ifndef __pinocchio_unsupported_hpp__
#define __pinocchio_unsupported_hpp__

#include "pinocchio/deprecation.hpp"

#define PINOCCHIO_UNSUPPORTED PINOCCHIO_DEPRECATED
#define PINOCCHIO_UNSUPPORTED_MESSAGE PINOCCHIO_DEPRECATED_MESSAGE

#endif // ifndef __pinocchio_unsupported_hpp__
