//
// Copyright (c) 2015-2024 CNRS INRIA
//
//
#ifndef __pinocchio_spatial_fcl_convertion_hpp__
#define __pinocchio_spatial_fcl_convertion_hpp__

#include "pinocchio/macros.hpp"

// clang-format off
PINOCCHIO_PRAGMA_DEPRECATED_HEADER(
  pinocchio/spatial/fcl-pinocchio-conversions.hpp,
  pinocchio/collision/fcl-pinocchio-conversions.hpp)
// clang-format on

#include "pinocchio/collision/fcl-pinocchio-conversions.hpp"

#endif // ifndef __pinocchio_spatial_fcl_convertion_hpp__
