//
// Copyright (c) 2021-2024 INRIA
//

#ifndef __pinocchio_algorithm_parallel_geometry_hpp__
#define __pinocchio_algorithm_parallel_geometry_hpp__

#include "pinocchio/macros.hpp"

// clang-format off
PINOCCHIO_PRAGMA_DEPRECATED_HEADER(pinocchio/algorithm/parallel/geometry.hpp, pinocchio/collision/parallel/geometry.hpp)
// clang-format on

#include "pinocchio/collision/parallel/geometry.hpp"

#endif // ifndef __pinocchio_algorithm_parallel_geometry_hpp__
