//
// Copyright (c) 2019 INRIA
//

#ifndef __pinocchio_parsers_sample_models_hpp__
#define __pinocchio_parsers_sample_models_hpp__

#include "pinocchio/macros.hpp"

// clang-format off
PINOCCHIO_PRAGMA_DEPRECATED_HEADER(pinocchio/parsers/sample-models.hpp, pinocchio/multibody/sample-models.hpp)
// clang-format on

#include "pinocchio/multibody/sample-models.hpp"

#endif // __pinocchio_parsers_sample_models_hpp__
