//
// Copyright (c) 2022 INRIA
//

#include "pinocchio/algorithm/proximal.hpp"

namespace pinocchio
{

  template struct PINOCCHIO_EXPLICIT_INSTANTIATION_DEFINITION_DLLAPI
    ProximalSettingsTpl<context::Scalar>;

} // namespace pinocchio
